/*************************************************************************************************************
 * (Abstract) OCF Base Interface module.
 *************************************************************************************************************/
(function() {
    /**
     * OCF base interface constructor.
     * 
     * @param {RegisterPacketParser} packetParser the register packet parser
     * @param {Function} sendPacket the send packet callback function
     */
    function OCFBaseInterface(packetParser, sendPacket) {
        this.packetParser = packetParser;
        this.sendPacket = sendPacket;
    }

    /**
     * Initialize the interface.
     * 
     * @param {*} info the necessary information to initialize this interface
     * @returns a promise
     */
    OCFBaseInterface.prototype.init = function(info) {
        return Promise.reject('');
    };

    /**
     * Reset the interface.
     * 
     * @returns a promise
     */
    OCFBaseInterface.prototype.reset = function() {
        return Promise.reject('');
    };

    /**
     * Handles the incoming packet.
     *
     * @param {*} packet the incoming packet
     * @returns true if handled, otherwise false
     */
    OCFBaseInterface.prototype.handlePacket = function(packet) {
        return false;
    };

    /**
     * A macro function to read register
     * @param {*} regInfo the register info
     * @return a promise
     */
    OCFBaseInterface.prototype.read = function(regInfo) {
        return Promise.reject('');
    };

    /**
     * A macro function to write register
     * @param {*} regInfo the register info
     * @param {*} data the data to write
     * @returns a promise
     */
    OCFBaseInterface.prototype.write = function(regInfo, data) {
        return Promise.reject('');
    };

    /* Exports the OCFBaseInterface object */
    window.OCFBaseInterface = OCFBaseInterface;
})();